@@ -17,21 +17,18 @@ from subprocess import (
 import subprocess
 import sys
 import threading
+from collections import OrderedDict
+from textwrap import dedent
 
 from git.compat import (
-    string_types,
     defenc,
     force_bytes,
-    PY3,
-    # just to satisfy flake8 on py3
-    unicode,
     safe_decode,
     is_posix,
     is_win,
 )
 from git.exc import CommandError
-from git.odict import OrderedDict
-from git.util import is_cygwin_git, cygpath
+from git.util import is_cygwin_git, cygpath, expand_path
 
 from .exc import (
     GitCommandError,
@@ -42,11 +39,10 @@ from .util import (
     stream_copy,
 )
 
-
-execute_kwargs = set(('istream', 'with_extended_output',
-                      'with_exceptions', 'as_process', 'stdout_as_string',
-                      'output_stream', 'with_stdout', 'kill_after_timeout',
-                      'universal_newlines', 'shell'))
+execute_kwargs = {'istream', 'with_extended_output',
+                  'with_exceptions', 'as_process', 'stdout_as_string',
+                  'output_stream', 'with_stdout', 'kill_after_timeout',
+                  'universal_newlines', 'shell', 'env', 'max_chunk_size'}
 
 log = logging.getLogger(__name__)
 log.addHandler(logging.NullHandler())
@@ -62,7 +58,7 @@ __all__ = ('Git',)
 
 def handle_process_output(process, stdout_handler, stderr_handler,
                           finalizer=None, decode_streams=True):
-    """Registers for notifications to lean that process output is ready to read, and dispatches lines to
+    """Registers for notifications to learn that process output is ready to read, and dispatches lines to
     the respective line handlers.
     This function returns once the finalizer returns
 
@@ -77,7 +73,7 @@ def handle_process_output(process, stdout_handler, stderr_handler,
         Set it to False if `universal_newline == True` (then streams are in text-mode)
         or if decoding must happen later (i.e. for Diffs).
     """
-    # Use 2 "pupm" threads and wait for both to finish.
+    # Use 2 "pump" threads and wait for both to finish.
     def pump_stream(cmdline, name, stream, is_decode, handler):
         try:
             for line in stream:
@@ -87,7 +83,7 @@ def handle_process_output(process, stdout_handler, stderr_handler,
                     handler(line)
         except Exception as ex:
             log.error("Pumping %r of cmd(%s) failed due to: %r", name, cmdline, ex)
-            raise CommandError(['<%s-pump>' % name] + cmdline, ex)
+            raise CommandError(['<%s-pump>' % name] + cmdline, ex) from ex
         finally:
             stream.close()
 
@@ -124,7 +120,7 @@ def dashify(string):
 
 
 def slots_to_dict(self, exclude=()):
-    return dict((s, getattr(self, s)) for s in self.__slots__ if s not in exclude)
+    return {s: getattr(self, s) for s in self.__slots__ if s not in exclude}
 
 
 def dict_to_slots_and__excluded_are_none(self, d, excluded=()):
@@ -142,8 +138,7 @@ CREATE_NO_WINDOW = 0x08000000
 ## CREATE_NEW_PROCESS_GROUP is needed to allow killing it afterwards,
 # see https://docs.python.org/3/library/subprocess.html#subprocess.Popen.send_signal
 PROC_CREATIONFLAGS = (CREATE_NO_WINDOW | subprocess.CREATE_NEW_PROCESS_GROUP
-                      if is_win and sys.version_info >= (2, 7)
-                      else 0)
+                      if is_win else 0)
 
 
 class Git(LazyMixin):
@@ -174,24 +169,140 @@ class Git(LazyMixin):
         dict_to_slots_and__excluded_are_none(self, d, excluded=self._excluded_)
 
     # CONFIGURATION
-    # The size in bytes read from stdout when copying git's output to another stream
-    max_chunk_size = io.DEFAULT_BUFFER_SIZE
 
     git_exec_name = "git"           # default that should work on linux and windows
 
     # Enables debugging of GitPython's git commands
     GIT_PYTHON_TRACE = os.environ.get("GIT_PYTHON_TRACE", False)
 
-    # Provide the full path to the git executable. Otherwise it assumes git is in the path
-    _git_exec_env_var = "GIT_PYTHON_GIT_EXECUTABLE"
-    GIT_PYTHON_GIT_EXECUTABLE = os.environ.get(_git_exec_env_var, git_exec_name)
-
     # If True, a shell will be used when executing git commands.
     # This should only be desirable on Windows, see https://github.com/gitpython-developers/GitPython/pull/126
     # and check `git/test_repo.py:TestRepo.test_untracked_files()` TC for an example where it is required.
     # Override this value using `Git.USE_SHELL = True`
     USE_SHELL = False
 
+    # Provide the full path to the git executable. Otherwise it assumes git is in the path
+    _git_exec_env_var = "GIT_PYTHON_GIT_EXECUTABLE"
+    _refresh_env_var = "GIT_PYTHON_REFRESH"
+    GIT_PYTHON_GIT_EXECUTABLE = None
+    # note that the git executable is actually found during the refresh step in
+    # the top level __init__
+
+    @classmethod
+    def refresh(cls, path=None):
+        """This gets called by the refresh function (see the top level
+        __init__).
+        """
+        # discern which path to refresh with
+        if path is not None:
+            new_git = os.path.expanduser(path)
+            new_git = os.path.abspath(new_git)
+        else:
+            new_git = os.environ.get(cls._git_exec_env_var, cls.git_exec_name)
+
+        # keep track of the old and new git executable path
+        old_git = cls.GIT_PYTHON_GIT_EXECUTABLE
+        cls.GIT_PYTHON_GIT_EXECUTABLE = new_git
+
+        # test if the new git executable path is valid
+
+        # - a GitCommandNotFound error is spawned by ourselves
+        # - a PermissionError is spawned if the git executable provided
+        #   cannot be executed for whatever reason
+        
+        has_git = False
+        try:
+            cls().version()
+            has_git = True
+        except (GitCommandNotFound, PermissionError):
+            pass
+
+        # warn or raise exception if test failed
+        if not has_git:
+            err = dedent("""\
+                Bad git executable.
+                The git executable must be specified in one of the following ways:
+                    - be included in your $PATH
+                    - be set via $%s
+                    - explicitly set via git.refresh()
+                """) % cls._git_exec_env_var
+
+            # revert to whatever the old_git was
+            cls.GIT_PYTHON_GIT_EXECUTABLE = old_git
+
+            if old_git is None:
+                # on the first refresh (when GIT_PYTHON_GIT_EXECUTABLE is
+                # None) we only are quiet, warn, or error depending on the
+                # GIT_PYTHON_REFRESH value
+
+                # determine what the user wants to happen during the initial
+                # refresh we expect GIT_PYTHON_REFRESH to either be unset or
+                # be one of the following values:
+                #   0|q|quiet|s|silence
+                #   1|w|warn|warning
+                #   2|r|raise|e|error
+
+                mode = os.environ.get(cls._refresh_env_var, "raise").lower()
+
+                quiet = ["quiet", "q", "silence", "s", "none", "n", "0"]
+                warn = ["warn", "w", "warning", "1"]
+                error = ["error", "e", "raise", "r", "2"]
+
+                if mode in quiet:
+                    pass
+                elif mode in warn or mode in error:
+                    err = dedent("""\
+                        %s
+                        All git commands will error until this is rectified.
+
+                        This initial warning can be silenced or aggravated in the future by setting the
+                        $%s environment variable. Use one of the following values:
+                            - %s: for no warning or exception
+                            - %s: for a printed warning
+                            - %s: for a raised exception
+
+                        Example:
+                            export %s=%s
+                        """) % (
+                        err,
+                        cls._refresh_env_var,
+                        "|".join(quiet),
+                        "|".join(warn),
+                        "|".join(error),
+                        cls._refresh_env_var,
+                        quiet[0])
+
+                    if mode in warn:
+                        print("WARNING: %s" % err)
+                    else:
+                        raise ImportError(err)
+                else:
+                    err = dedent("""\
+                        %s environment variable has been set but it has been set with an invalid value.
+
+                        Use only the following values:
+                            - %s: for no warning or exception
+                            - %s: for a printed warning
+                            - %s: for a raised exception
+                        """) % (
+                        cls._refresh_env_var,
+                        "|".join(quiet),
+                        "|".join(warn),
+                        "|".join(error))
+                    raise ImportError(err)
+
+                # we get here if this was the init refresh and the refresh mode
+                # was not error, go ahead and set the GIT_PYTHON_GIT_EXECUTABLE
+                # such that we discern the difference between a first import
+                # and a second import
+                cls.GIT_PYTHON_GIT_EXECUTABLE = cls.git_exec_name
+            else:
+                # after the first refresh (when GIT_PYTHON_GIT_EXECUTABLE
+                # is no longer None) we raise an exception
+                raise GitCommandNotFound("git", err)
+
+        return has_git
+
     @classmethod
     def is_cygwin(cls):
         return is_cygwin_git(cls.GIT_PYTHON_GIT_EXECUTABLE)
@@ -210,6 +321,9 @@ class Git(LazyMixin):
             but git stops liking them as it will escape the backslashes.
             Hence we undo the escaping just to be sure.
             """
+            url = os.path.expandvars(url)
+            if url.startswith('~'):
+                url = os.path.expanduser(url)
             url = url.replace("\\\\", "\\").replace("\\", "/")
 
         return url
@@ -242,8 +356,11 @@ class Git(LazyMixin):
                 proc.stderr.close()
 
             # did the process finish already so we have a return code ?
-            if proc.poll() is not None:
-                return
+            try:
+                if proc.poll() is not None:
+                    return
+            except OSError as ex:
+                log.info("Ignored error after process had died: %r", ex)
 
             # can be that nothing really exists anymore ...
             if os is None or getattr(os, 'kill', None) is None:
@@ -254,14 +371,14 @@ class Git(LazyMixin):
                 proc.terminate()
                 proc.wait()    # ensure process goes away
             except OSError as ex:
-                log.info("Ignored error after process has dies: %r", ex)
-                pass  # ignore error when process already died
+                log.info("Ignored error after process had died: %r", ex)
             except AttributeError:
                 # try windows
                 # for some reason, providing None for stdout/stderr still prints something. This is why
                 # we simply use the shell and redirect to nul. Its slower than CreateProcess, question
                 # is whether we really want to see all these messages. Its annoying no matter what.
-                call(("TASKKILL /F /T /PID %s 2>nul 1>nul" % str(proc.pid)), shell=True)
+                if is_win:
+                    call(("TASKKILL /F /T /PID %s 2>nul 1>nul" % str(proc.pid)), shell=True)
             # END exception handling
 
         def __getattr__(self, attr):
@@ -275,7 +392,7 @@ class Git(LazyMixin):
             :raise GitCommandError: if the return status is not 0"""
             if stderr is None:
                 stderr = b''
-            stderr = force_bytes(stderr)
+            stderr = force_bytes(data=stderr, encoding='utf-8')
 
             status = self.proc.wait()
 
@@ -359,10 +476,10 @@ class Git(LazyMixin):
 
         def readlines(self, size=-1):
             if self._nbr == self._size:
-                return list()
+                return []
 
             # leave all additional logic to our readline method, we just check the size
-            out = list()
+            out = []
             nbr = 0
             while True:
                 line = self.readline()
@@ -377,8 +494,12 @@ class Git(LazyMixin):
             # END readline loop
             return out
 
+        # skipcq: PYL-E0301
         def __iter__(self):
             return self
+        
+        def __next__(self):
+            return self.next()
 
         def next(self):
             line = self.readline()
@@ -404,7 +525,7 @@ class Git(LazyMixin):
            It is meant to be the working tree directory if available, or the
            .git directory in case of bare repositories."""
         super(Git, self).__init__()
-        self._working_dir = working_dir
+        self._working_dir = expand_path(working_dir)
         self._git_options = ()
         self._persistent_git_options = []
 
@@ -470,6 +591,8 @@ class Git(LazyMixin):
                 with_stdout=True,
                 universal_newlines=False,
                 shell=None,
+                env=None,
+                max_chunk_size=io.DEFAULT_BUFFER_SIZE,
                 **subprocess_kwargs
                 ):
         """Handles executing the command on the shell and consumes and returns
@@ -513,6 +636,14 @@ class Git(LazyMixin):
             decoded into a string using the default encoding (usually utf-8).
             The latter can fail, if the output contains binary data.
 
+        :param env:
+            A dictionary of environment variables to be passed to `subprocess.Popen`.
+            
+        :param max_chunk_size:
+            Maximum number of bytes in one chunk of data passed to the output_stream in
+            one invocation of write() method. If the given number is not positive then
+            the default value is used.
+
         :param subprocess_kwargs:
             Keyword arguments to be passed to subprocess.Popen. Please note that
             some of the valid kwargs are already set by this method, the ones you
@@ -539,7 +670,7 @@ class Git(LazyMixin):
             * str(output) if extended_output = False (Default)
             * tuple(int(status), str(stdout), str(stderr)) if extended_output = True
 
-            if ouput_stream is True, the stdout value will be your output stream:
+            if output_stream is True, the stdout value will be your output stream:
             * output_stream if extended_output = False
             * tuple(int(status), output_stream, str(stderr)) if extended_output = True
 
@@ -558,6 +689,7 @@ class Git(LazyMixin):
         cwd = self._working_dir or os.getcwd()
 
         # Start the process
+        inline_env = env
         env = os.environ.copy()
         # Attempt to force all output to plain ascii english, which is what some parsing code
         # may expect.
@@ -566,6 +698,8 @@ class Git(LazyMixin):
         env["LANGUAGE"] = "C"
         env["LC_ALL"] = "C"
         env.update(self._environment)
+        if inline_env is not None:
+            env.update(inline_env)
 
         if is_win:
             cmd_not_found_exception = OSError
@@ -580,9 +714,12 @@ class Git(LazyMixin):
 
         stdout_sink = (PIPE
                        if with_stdout
-                       else getattr(subprocess, 'DEVNULL', open(os.devnull, 'wb')))
-        log.debug("Popen(%s, cwd=%s, universal_newlines=%s, shell=%s)",
-                  command, cwd, universal_newlines, shell)
+                       else getattr(subprocess, 'DEVNULL', None) or open(os.devnull, 'wb'))
+        istream_ok = "None"
+        if istream:
+            istream_ok = "<valid stream>"
+        log.debug("Popen(%s, cwd=%s, universal_newlines=%s, shell=%s, istream=%s)",
+                  command, cwd, universal_newlines, shell, istream_ok)
         try:
             proc = Popen(command,
                          env=env,
@@ -598,7 +735,7 @@ class Git(LazyMixin):
                          **subprocess_kwargs
                          )
         except cmd_not_found_exception as err:
-            raise GitCommandNotFound(command, err)
+            raise GitCommandNotFound(command, err) from err
 
         if as_process:
             return self.AutoInterrupt(proc, command)
@@ -638,6 +775,7 @@ class Git(LazyMixin):
         status = 0
         stdout_value = b''
         stderr_value = b''
+        newline = "\n" if universal_newlines else b"\n"
         try:
             if output_stream is None:
                 if kill_after_timeout:
@@ -646,20 +784,23 @@ class Git(LazyMixin):
                 if kill_after_timeout:
                     watchdog.cancel()
                     if kill_check.isSet():
-                        stderr_value = 'Timeout: the command "%s" did not complete in %d ' \
-                                       'secs.' % (" ".join(command), kill_after_timeout)
+                        stderr_value = ('Timeout: the command "%s" did not complete in %d '
+                                        'secs.' % (" ".join(command), kill_after_timeout))
+                        if not universal_newlines:
+                            stderr_value = stderr_value.encode(defenc)
                 # strip trailing "\n"
-                if stdout_value.endswith(b"\n"):
+                if stdout_value.endswith(newline):
                     stdout_value = stdout_value[:-1]
-                if stderr_value.endswith(b"\n"):
+                if stderr_value.endswith(newline):
                     stderr_value = stderr_value[:-1]
                 status = proc.returncode
             else:
-                stream_copy(proc.stdout, output_stream, self.max_chunk_size)
-                stdout_value = output_stream
+                max_chunk_size = max_chunk_size if max_chunk_size and max_chunk_size > 0 else io.DEFAULT_BUFFER_SIZE
+                stream_copy(proc.stdout, output_stream, max_chunk_size)
+                stdout_value = proc.stdout.read()
                 stderr_value = proc.stderr.read()
                 # strip trailing "\n"
-                if stderr_value.endswith(b"\n"):
+                if stderr_value.endswith(newline):
                     stderr_value = stderr_value[:-1]
                 status = proc.wait()
             # END stdout handling
@@ -747,7 +888,7 @@ class Git(LazyMixin):
         if len(name) == 1:
             if value is True:
                 return ["-%s" % name]
-            elif type(value) is not bool:
+            elif value not in (False, None):
                 if split_single_char_options:
                     return ["-%s" % name, "%s" % value]
                 else:
@@ -755,13 +896,13 @@ class Git(LazyMixin):
         else:
             if value is True:
                 return ["--%s" % dashify(name)]
-            elif type(value) is not bool:
+            elif value is not False and value is not None:
                 return ["--%s=%s" % (dashify(name), value)]
         return []
 
     def transform_kwargs(self, split_single_char_options=True, **kwargs):
         """Transforms Python style kwargs into git command line options."""
-        args = list()
+        args = []
         kwargs = OrderedDict(sorted(kwargs.items(), key=lambda x: x[0]))
         for k, v in kwargs.items():
             if isinstance(v, (list, tuple)):
@@ -774,18 +915,12 @@ class Git(LazyMixin):
     @classmethod
     def __unpack_args(cls, arg_list):
         if not isinstance(arg_list, (list, tuple)):
-            # This is just required for unicode conversion, as subprocess can't handle it
-            # However, in any other case, passing strings (usually utf-8 encoded) is totally fine
-            if not PY3 and isinstance(arg_list, unicode):
-                return [arg_list.encode(defenc)]
             return [str(arg_list)]
 
-        outlist = list()
+        outlist = []
         for arg in arg_list:
             if isinstance(arg_list, (list, tuple)):
                 outlist.extend(cls.__unpack_args(arg))
-            elif not PY3 and isinstance(arg_list, unicode):
-                outlist.append(arg_list.encode(defenc))
             # END recursion
             else:
                 outlist.append(str(arg))
@@ -822,27 +957,30 @@ class Git(LazyMixin):
             is realized as non-existent
 
         :param kwargs:
-            is a dict of keyword arguments.
-            This function accepts the same optional keyword arguments
-            as execute().
+            It contains key-values for the following:
+            - the :meth:`execute()` kwds, as listed in :var:`execute_kwargs`;
+            - "command options" to be converted by :meth:`transform_kwargs()`;
+            - the `'insert_kwargs_after'` key which its value must match one of ``*args``,
+              and any cmd-options will be appended after the matched arg.
+
+        Examples::
 
-        ``Examples``::
             git.rev_list('master', max_count=10, header=True)
 
+        turns into::
+
+           git rev-list max-count 10 --header master
+
         :return: Same as ``execute``"""
         # Handle optional arguments prior to calling transform_kwargs
         # otherwise these'll end up in args, which is bad.
-        _kwargs = dict()
-        for kwarg in execute_kwargs:
-            try:
-                _kwargs[kwarg] = kwargs.pop(kwarg)
-            except KeyError:
-                pass
+        exec_kwargs = {k: v for k, v in kwargs.items() if k in execute_kwargs}
+        opts_kwargs = {k: v for k, v in kwargs.items() if k not in execute_kwargs}
 
-        insert_after_this_arg = kwargs.pop('insert_kwargs_after', None)
+        insert_after_this_arg = opts_kwargs.pop('insert_kwargs_after', None)
 
         # Prepare the argument list
-        opt_args = self.transform_kwargs(**kwargs)
+        opt_args = self.transform_kwargs(**opts_kwargs)
         ext_args = self.__unpack_args([a for a in args if a is not None])
 
         if insert_after_this_arg is None:
@@ -850,12 +988,12 @@ class Git(LazyMixin):
         else:
             try:
                 index = ext_args.index(insert_after_this_arg)
-            except ValueError:
-                raise ValueError("Couldn't find argument '%s' in args %s to insert kwargs after"
-                                 % (insert_after_this_arg, str(ext_args)))
+            except ValueError as err:
+                raise ValueError("Couldn't find argument '%s' in args %s to insert cmd options after"
+                                 % (insert_after_this_arg, str(ext_args))) from err
             # end handle error
             args = ext_args[:index + 1] + opt_args + ext_args[index + 1:]
-        # end handle kwargs
+        # end handle opts_kwargs
 
         call = [self.GIT_PYTHON_GIT_EXECUTABLE]
 
@@ -870,7 +1008,7 @@ class Git(LazyMixin):
         call.append(dashify(method))
         call.extend(args)
 
-        return self.execute(call, **_kwargs)
+        return self.execute(call, **exec_kwargs)
 
     def _parse_object_header(self, header_line):
         """
@@ -900,7 +1038,7 @@ class Git(LazyMixin):
         if isinstance(ref, bytes):
             # Assume 40 bytes hexsha - bin-to-ascii for some reason returns bytes, not text
             refstr = ref.decode('ascii')
-        elif not isinstance(ref, string_types):
+        elif not isinstance(ref, str):
             refstr = str(ref)               # could be ref-object
 
         if not refstr.endswith("\n"):
