@@ -24,7 +24,6 @@ from .util import (
     parse_actor_and_date,
     from_timestamp,
 )
-from git.compat import text_type
 
 from time import (
     time,
@@ -81,7 +80,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
         :param tree: Tree
             Tree object
         :param author: Actor
-            is the author string ( will be implicitly converted into an Actor object )
+            is the author Actor object
         :param authored_date: int_seconds_since_epoch
             is the authored DateTime - use time.gmtime() to convert it into a
             different format
@@ -137,10 +136,45 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
     def _get_intermediate_items(cls, commit):
         return commit.parents
 
+    @classmethod
+    def _calculate_sha_(cls, repo, commit):
+        '''Calculate the sha of a commit.
+
+        :param repo: Repo object the commit should be part of
+        :param commit: Commit object for which to generate the sha
+        '''
+
+        stream = BytesIO()
+        commit._serialize(stream)
+        streamlen = stream.tell()
+        stream.seek(0)
+
+        istream = repo.odb.store(IStream(cls.type, streamlen, stream))
+        return istream.binsha
+
+    def replace(self, **kwargs):
+        '''Create new commit object from existing commit object.
+
+        Any values provided as keyword arguments will replace the
+        corresponding attribute in the new object.
+        '''
+
+        attrs = {k: getattr(self, k) for k in self.__slots__}
+
+        for attrname in kwargs:
+            if attrname not in self.__slots__:
+                raise ValueError('invalid attribute name')
+
+        attrs.update(kwargs)
+        new_commit = self.__class__(self.repo, self.NULL_BIN_SHA, **attrs)
+        new_commit.binsha = self._calculate_sha_(self.repo, new_commit)
+
+        return new_commit
+
     def _set_cache_(self, attr):
         if attr in Commit.__slots__:
             # read the data in a chunk, its faster - then provide a file wrapper
-            binsha, typename, self.size, stream = self.repo.odb.stream(self.binsha)  # @UnusedVariable
+            _binsha, _typename, self.size, stream = self.repo.odb.stream(self.binsha)
             self._deserialize(BytesIO(stream.read()))
         else:
             super(Commit, self)._set_cache_(attr)
@@ -174,8 +208,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
         # as the empty paths version will ignore merge commits for some reason.
         if paths:
             return len(self.repo.git.rev_list(self.hexsha, '--', paths, **kwargs).splitlines())
-        else:
-            return len(self.repo.git.rev_list(self.hexsha, **kwargs).splitlines())
+        return len(self.repo.git.rev_list(self.hexsha, **kwargs).splitlines())
 
     @property
     def name_rev(self):
@@ -271,7 +304,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
             # END handle extra info
 
             assert len(hexsha) == 40, "Invalid line: %s" % hexsha
-            yield Commit(repo, hex_to_bin(hexsha))
+            yield cls(repo, hex_to_bin(hexsha))
         # END for each line in stream
         # TODO: Review this - it seems process handling got a bit out of control
         # due to many developers trying to fix the open file handles issue
@@ -316,7 +349,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
                 parent_commits = [repo.head.commit]
             except ValueError:
                 # empty repositories have no head commit
-                parent_commits = list()
+                parent_commits = []
             # END handle parent commits
         else:
             for p in parent_commits:
@@ -377,13 +410,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
                          committer, committer_time, committer_offset,
                          message, parent_commits, conf_encoding)
 
-        stream = BytesIO()
-        new_commit._serialize(stream)
-        streamlen = stream.tell()
-        stream.seek(0)
-
-        istream = repo.odb.store(IStream(cls.type, streamlen, stream))
-        new_commit.binsha = istream.binsha
+        new_commit.binsha = cls._calculate_sha_(repo, new_commit)
 
         if head:
             # need late import here, importing git at the very beginning throws
@@ -437,7 +464,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
         write(b"\n")
 
         # write plain bytes, be sure its encoded according to our encoding
-        if isinstance(self.message, text_type):
+        if isinstance(self.message, str):
             write(self.message.encode(self.encoding))
         else:
             write(self.message)
@@ -450,7 +477,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
         readline = stream.readline
         self.tree = Tree(self.repo, hex_to_bin(readline().split()[1]), Tree.tree_id << 12, '')
 
-        self.parents = list()
+        self.parents = []
         next_line = None
         while True:
             parent_line = readline()
@@ -484,7 +511,8 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
         buf = enc.strip()
         while buf:
             if buf[0:10] == b"encoding ":
-                self.encoding = buf[buf.find(' ') + 1:].decode('ascii')
+                self.encoding = buf[buf.find(' ') + 1:].decode(
+                    self.encoding, 'ignore')
             elif buf[0:7] == b"gpgsig ":
                 sig = buf[buf.find(b' ') + 1:] + b"\n"
                 is_next_header = False
@@ -498,7 +526,7 @@ class Commit(base.Object, Iterable, Diffable, Traversable, Serializable):
                         break
                     sig += sigbuf[1:]
                 # end read all signature
-                self.gpgsig = sig.rstrip(b"\n").decode('ascii')
+                self.gpgsig = sig.rstrip(b"\n").decode(self.encoding, 'ignore')
                 if is_next_header:
                     continue
             buf = readline().strip()
